# Copyright 2021 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from rc_django.views import RestView
from rc_django.models import RestProxy
from django.template import loader, TemplateDoesNotExist
from django.urls import reverse
from django.http import HttpResponse, HttpResponseRedirect
from userservice.user import UserService
from urllib.parse import quote, unquote, urlencode, urlparse, parse_qs
from base64 import b64encode
from logging import getLogger
import re

logger = getLogger(__name__)


class RestProxyView(RestView):
    template_name = "proxy.html"

    @staticmethod
    def format_search_params(url):
        params = {}
        query_params = parse_qs(urlparse(url).query)
        for param in query_params:
            params[param] = ",".join(query_params[param])
        return params

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        service = kwargs.get("service")
        url = kwargs.get("url")
        headers = kwargs.get("headers", {})
        service_name = service
        use_pre = False
        is_image = False
        user_service = UserService()

        if service == "iasystem":
            headers["Accept"] = "application/vnd.collection+json"
            service_name = 'iasystem_uw'
        elif service == "sws" or service == "gws":
            headers["X-UW-Act-as"] = user_service.get_original_user()
        elif service == "calendar":
            use_pre = True

        proxy = RestProxy(service_name)
        response = proxy.get_api_response(url, headers)

        if (response.status == 200 and
                re.match(r"/idcard/v1/photo/[0-9A-F]{32}", url)):
            # Handle known images
            is_image = True
            content = b64encode(response.data).decode("utf-8")
        elif use_pre:
            content = response.data
        else:
            content = proxy.formatted

        context.update({
            "url": unquote(url),
            "content": content,
            "json_data": proxy.json,
            "response_code": response.status,
            "time_taken": "{:f} seconds".format(proxy.duration),
            "headers": response.headers,
            "override_user": user_service.get_override_user(),
            "use_pre": use_pre,
            "is_image": is_image,
        })

        try:
            loader.get_template("restclients/extra_info.html")
            context["has_extra_template"] = True
            context["extra_template"] = "restclients/extra_info.html"
        except TemplateDoesNotExist:
            pass

        try:
            search_template_path = re.sub(r"[.?].*$", "", url)
            search_template = "proxy/{}{}.html".format(service,
                                                       search_template_path)
            loader.get_template(search_template)
            context["search_template"] = search_template
            context["search"] = self.format_search_params(url)
        except TemplateDoesNotExist:
            context["search_template"] = None

        return context

    def get(self, request, *args, **kwargs):
        """
        Fetch an API resource and render it, formatted for a browser.
        """
        # Using args for these URLs for backwards-compatibility
        kwargs["service"] = args[0]
        kwargs["url"] = "/" + (args[1] if len(args) > 1 else "")

        if request.GET:
            kwargs["url"] += "?" + urlencode(request.GET)
        else:
            try:
                path, qs = kwargs["url"].split("?")
                kwargs["url"] = "?".join([quote(path), qs])
            except ValueError:
                pass

        try:
            context = self.get_context_data(**kwargs)
        except (AttributeError, ImportError):
            return HttpResponse(
                "Missing service: {}".format(kwargs["service"]), status=404)

        return self.render_to_response(context)


class RestSearchView(RestView):
    template_name = "customform.html"
    form_action_url = "restclients_customform"

    def get_context_data(self, **kwargs):
        service = kwargs.get("service")
        path = kwargs.get("path", "")

        context = super().get_context_data(**kwargs)

        form_path = "customform/{}/{}".format(service, path)
        try:
            loader.get_template("restclients/{}".format(form_path))
            context["form_template"] = "restclients/{}".format(form_path)
        except TemplateDoesNotExist:
            loader.get_template(form_path)
            context["form_template"] = form_path

        context["form_action"] = reverse(self.form_action_url, args=[
            service, path.replace(".html", "")])
        return context

    def get(self, request, *args, **kwargs):
        """
        Renders a custom form for searching a REST service.
        """
        # Using args for these URLs for backwards-compatibility
        kwargs["service"] = args[0]
        kwargs["path"] = args[1] if len(args) > 1 else ""

        try:
            context = self.get_context_data(**kwargs)
        except TemplateDoesNotExist as ex:
            return HttpResponse("Missing template: {}".format(ex), status=404)

        return self.render_to_response(context)

    def post(self, request, *args, **kwargs):
        """
        Entry point for custom search form submissions.
        Convert form data to actual API urls, and redirect to proxy view.
        """
        service = args[0]
        url = args[1] if len(args) > 1 else ""

        try:
            service, url, params = self.get_proxy_url(request, service, url)
        except KeyError as ex:
            return HttpResponse("Missing reqired form value: {}".format(ex),
                                status=400)

        url = reverse("restclients_proxy", args=[service, url])
        if params:
            url += "?" + urlencode(params)
        return HttpResponseRedirect(url)

    def format_params(self, request):
        return {k: v for k, v in request.POST.items() if (
            k != "csrfmiddlewaretoken")}

    def get_proxy_url(self, request, service, url):
        params = None
        if service == "libcurrics":
            if "course" == url:
                url = "currics_db/api/v1/data/{}/{}/{}/{}/{}/{}".format(
                    "course",
                    request.POST["year"],
                    request.POST["quarter"],
                    request.POST["curriculum_abbr"],
                    request.POST["course_number"],
                    request.POST["section_id"])
            elif "default" == url:
                url = "currics_db/api/v1/data/defaultGuide/{}".format(
                    request.POST["campus"])
        elif service == "libraries":
            url = "mylibinfo/v1/"
            params = self.format_params(request)

        return service, url, params



# Copyright 2021 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0


def can_proxy_restclient(request, service, url):
    if service:
        return service not in ['secret']
    return True



# Copyright 2021 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from rc_django.views import RestView
from rc_django.models import DegradePerformance


class DegradePerformanceView(RestView):
    template_name = "cause_errors.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        problem_str = kwargs.get("problem")
        problems = DegradePerformance(serialized=problem_str)

        context["errors"] = []
        for service in problems.services():
            context["errors"].append({
                "name": service,
                "status": problems.get_status(service),
                "content": problems.get_content(service),
                "load_time": problems.get_load_time(service),
            })
        return context

    def get(self, request, *args, **kwargs):
        kwargs["problem"] = request.session.get("RESTCLIENTS_ERRORS")
        context = self.get_context_data(**kwargs)
        return self.render_to_response(context)

    def post(self, request, *args, **kwargs):
        problem_str = request.session.get("RESTCLIENTS_ERRORS")
        problems = DegradePerformance(serialized=problem_str)

        for key in problems.services():
            keepit = "keep_{}".format(key)
            if keepit not in request.POST:
                problems.remove_service(key)
            else:
                problems.set_status(
                    key, request.POST.get("{}_status".format(key), None))
                problems.set_content(
                    key, request.POST.get("{}_content".format(key), None))
                problems.set_load_time(
                    key, request.POST.get("{}_load_time".format(key), None))

        new_service = request.POST.get("new_service_name", None)
        if new_service:
            key = request.POST["new_service_name"]
            problems.set_status(
                key, request.POST.get("new_service_status", None))
            problems.set_content(
                key, request.POST.get("new_service_content", None))
            problems.set_load_time(
                key, request.POST.get("new_service_load_time", None))

        request.session["RESTCLIENTS_ERRORS"] = problems.serialize()
        kwargs["problem"] = request.session["RESTCLIENTS_ERRORS"]

        context = self.get_context_data(**kwargs)
        return self.render_to_response(context)



# Copyright 2021 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from django.db import models
from django.urls import reverse
from restclients_core.dao import DAO
from restclients_core.models import MockHTTP
from restclients_core.exceptions import DataFailureException
from time import time
import json
import re


class RestProxy():
    def __init__(self, service):
        self.service = service
        self.response = None
        self._request_start = 0
        self._request_end = 0

    @property
    def dao(self):
        def get_all_subclasses(base):
            return base.__subclasses__() + [g for s in base.__subclasses__()
                                            for g in get_all_subclasses(s)]

        for subclass in get_all_subclasses(DAO):
            dao = subclass()
            if dao.service_name() == self.service:
                return dao
        raise ImportError()

    @property
    def duration(self):
        return self._request_end - self._request_start

    @property
    def json(self):
        try:
            return json.dumps(json.loads(self.response.data), sort_keys=True)
        except ValueError:
            pass

    @property
    def formatted(self):
        try:
            # Assume json, and try to format it.
            return self.format_json()
        except ValueError:
            return self.format_html()

    def get_api_response(self, url, headers={}):
        self._request_start = time()
        try:
            response = self.dao.getURL(url, headers)
        except DataFailureException as ex:
            response = MockHTTP()
            response.status = ex.status
            response.data = ex.msg
        self._request_end = time()
        self.response = response
        return self.response

    def format_json(self):
        data = json.loads(self.response.data)
        formatted = json.dumps(data, sort_keys=True, indent=4)
        formatted = formatted.replace("&", "&amp;")
        formatted = formatted.replace("<", "&lt;")
        formatted = formatted.replace(">", "&gt;")
        formatted = formatted.replace(" ", "&nbsp;")
        formatted = formatted.replace("\n", "<br/>\n")

        base_url = reverse("restclients_proxy", args=["xx", "xx"])
        base_url = base_url.replace('/xx/xx', '')

        formatted = re.sub(r"\"/(.*?)\"",
                           r'"<a href="{}/{}/\1">/\1</a>"'.format(
                               base_url, self.service),
                           formatted)
        return formatted

    def format_html(self):
        try:
            content = self.response.data.decode('utf-8')
        except AttributeError:
            content = self.response.data

        base_url = reverse("restclients_proxy", args=["xx", "xx"])
        base_url = base_url.replace('/xx/xx', '')

        formatted = re.sub(r"href\s*=\s*[\"\']/(.*?)[\"\']",
                           r'href="{}/{}/\1"'.format(base_url, self.service),
                           content, flags=re.I)
        formatted = re.sub(
            re.compile(r"<style.*/style>", flags=re.S | re.I), "", formatted)
        formatted = self.clean_self_closing_divs(formatted)
        return formatted

    @staticmethod
    def clean_self_closing_divs(content):
        return re.sub(r"((<div[^>]*?)/>)",
                      r"<!-- \g<1> -->\g<2>></div>",
                      content)


class DegradePerformance(object):
    def __init__(self, serialized=None):
        self.problems = {}

        if serialized:
            self.problems = json.loads(serialized)

    def remove_service(self, service):
        del self.problems[service]

    def set_status(self, service, value):
        self._set(service, "status", value)

    def set_content(self, service, value):
        self._set(service, "content", value)

    def set_load_time(self, service, value):
        self._set(service, "load_time", value)

    def get_status(self, service):
        return self._get(service, "status")

    def get_content(self, service):
        return self._get(service, "content")

    def get_load_time(self, service):
        return self._get(service, "load_time")

    def services(self):
        return self.problems.keys()

    def _set(self, service, key, value):
        self._add_service(service)
        self.problems[service][key] = value

    def _get(self, service, key):
        if service in self.problems:
            if key in self.problems[service]:
                return self.problems[service][key]
        return None

    def _add_service(self, service):
        if service not in self.problems:
            self.problems[service] = {}

    def serialize(self):
        return json.dumps(self.problems)



# Copyright 2021 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from django.views.generic.base import TemplateView
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_protect
from django.template import loader, TemplateDoesNotExist
from rc_django.decorators import restclient_admin_required


@method_decorator(csrf_protect, name='dispatch')
@method_decorator(restclient_admin_required, name='dispatch')
class RestView(TemplateView):
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        try:
            loader.get_template("restclients/proxy_wrapper.html")
            context["wrapper_template"] = "restclients/proxy_wrapper.html"
        except TemplateDoesNotExist:
            context["wrapper_template"] = "proxy_wrapper.html"
        return context



# Copyright 2021 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from django.test import TestCase
from rc_django.models import RestProxy
from rc_django.tests.test_views import TEST_DAO, SUB_DAO
from restclients_core.models import MockHTTP


class RestProxyTest(TestCase):
    def setUp(self):
        response = MockHTTP()
        response.status = 200
        response.data = "ok"

        self.proxy = RestProxy("pws")
        self.proxy.response = response

    def test_dao_instance(self):
        self.assertEqual(type(RestProxy("test").dao), TEST_DAO)
        self.assertEquals(type(RestProxy("test_sub").dao), SUB_DAO)

        # Missing service
        proxy = RestProxy("fake")
        self.assertRaises(ImportError, getattr, proxy, "dao")

    def test_get_mock_response(self):
        proxy = RestProxy("test_sub")
        response = proxy.get_api_response("/foo")
        self.assertEqual(response.status, 200)
        self.assertEqual(response.data, "ok")
        self.assertTrue(isinstance(proxy.duration, float))

    def test_simple(self):
        closed = "<div/>"
        valid = "<!-- <div/> --><div></div>"

        self.assertEquals(valid, self.proxy.clean_self_closing_divs(closed))

    def test_2_simple(self):
        closed = '<div/><div id="1"/>'
        valid = ('<!-- <div/> --><div></div>'
                 '<!-- <div id="1"/> --><div id="1"></div>')

        self.assertEqual(valid, self.proxy.clean_self_closing_divs(closed))

    def test_valid_div(self):
        valid = "<div id='test_id'></div>"
        self.assertEqual(valid, self.proxy.clean_self_closing_divs(valid))

    def test_div_then_valid_self_closing(self):
        valid = "<div id='test_id'></div><br/>"
        self.assertEqual(valid, self.proxy.clean_self_closing_divs(valid))

    def test_json(self):
        content = '{"Href": "/identity/v2/entity.json"}'
        self.proxy.response.data = content
        self.assertEqual(self.proxy.json, content)

        content = '{"Decimal": 5.678}'
        self.proxy.response.data = content
        self.assertEqual(self.proxy.json, content)

    def test_format_json(self):
        json_data = '{"Href": "/identity/v2/entity.json"}'
        formatted = (u'{<br/>\n&nbsp;&nbsp;&nbsp;&nbsp;"Href":&nbsp;'
                     u'"<a href="/view/pws/identity/v2/entity.json">'
                     u'/identity/v2/entity.json</a>"<br/>\n}')
        self.proxy.response.data = json_data
        self.assertEqual(self.proxy.format_json(), formatted)

        json_data = '{"Decimal": 5.678}'
        formatted = ('{<br/>\n&nbsp;&nbsp;&nbsp;&nbsp;"Decimal":'
                     '&nbsp;5.678<br/>\n}')
        self.proxy.response.data = json_data
        self.assertEqual(self.proxy.format_json(), formatted)

        self.proxy.response.data = "<p></p>"
        self.assertRaises(ValueError, self.proxy.format_json)

    def test_format_html(self):
        output = '<a href="/view/pws/api/v1/test"></a>'

        self.proxy.response.data = '<a href="/api/v1/test"></a>'
        self.assertEqual(self.proxy.format_html(), output)

        self.proxy.response.data = '<a HREF="/api/v1/test"></a>'
        self.assertEqual(self.proxy.format_html(), output)

        # Binary string
        self.proxy.response.data = b'<a href="/api/v1/test"></a>'
        self.assertEqual(self.proxy.format_html(), output)

        # Single quotes
        self.proxy.response.data = "<a href='/api/v1/test'></a>"
        self.assertEqual(self.proxy.format_html(), output)

        # Style tags
        self.proxy.response.data = (
            '<style>h1 {color:red;}</style><a href="/api/v1/test"></a>')
        self.assertEqual(self.proxy.format_html(), output)

        self.proxy.response.data = (
            b'<STYLE>h1 {color:red;}</STYLE><a href="/api/v1/test"></a>')
        self.assertEqual(self.proxy.format_html(), output)



# Copyright 2021 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

from django.urls import re_path
from rc_django.views.errors import DegradePerformanceView
from rc_django.views.rest_proxy import RestSearchView, RestProxyView

urlpatterns = [
    re_path(r'^errors',
            DegradePerformanceView.as_view(), name="restclients_errors"),
    re_path(r'^search/(\w+)/(.*)$',
            RestSearchView.as_view(), name="restclients_customform"),
    re_path(r'^view/(\w+)/(.*)$',
            RestProxyView.as_view(), name="restclients_proxy"),
]



# Copyright 2021 UW-IT, University of Washington
# SPDX-License-Identifier: Apache-2.0

# -*- coding: utf-8 -*-
from django.test import TestCase
from django.conf import settings
from django.test.utils import override_settings
from django.contrib.auth.models import User
from django.urls import reverse
from restclients_core.dao import DAO, MockDAO
from restclients_core.models import MockHTTP
from rc_django.views.rest_proxy import RestSearchView, RestProxyView


class TEST_DAO(DAO):
    def __init__(self):
        super(TEST_DAO, self).__init__()

    def service_name(self):
        return "test"

    def get_default_service_setting(self, key):
        if "DAO_CLASS" == key:
            return "rc_django.tests.test_views.Backend"


class SUB_DAO(TEST_DAO):
    def __init__(self):
        super(SUB_DAO, self).__init__()

    def service_name(self):
        return "test_sub"


class Backend(MockDAO):
    def load(self, method, url, headers, body):
        response = MockHTTP()
        response.status = 200
        response.data = "ok"
        return response


def missing_url(name, *args, **kwargs):
    try:
        url = reverse(name, *args, **kwargs)
    except Exception as ex:
        print(str(ex))
        if getattr(settings, "RESTCLIENTS_REQUIRE_VIEW_TESTS", False):
            raise
        return True

    return False


def get_user(username):
    try:
        user = User.objects.get(username=username)
        return user
    except Exception as ex:
        user = User.objects.create_user(username, password='pass')
        return user


def get_user_pass(username):
    return 'pass'


class ViewNoAuthTest(TestCase):
    def test_service_errors(self):
        get_user('test_view')
        self.client.login(
            username='test_view', password=get_user_pass('test_view'))

        # No auth module in settings
        url = reverse("restclients_proxy", args=["test", "test/v1"])
        response = self.client.get(url)
        self.assertEquals(response.status_code, 401)

        url = reverse("restclients_customform", args=["test", "index.html"])
        response = self.client.get(url)
        self.assertEquals(response.status_code, 401)


@override_settings(
    RESTCLIENTS_ADMIN_AUTH_MODULE='rc_django.tests.can_proxy_restclient')
class RestProxyViewTest(TestCase):
    def test_bad_url(self):
        # Something was sending urls that should have been
        # ...=&reg_id=... into ...=®_id=A
        # That should be fixed, but in the mean time we shouldn't crash
        url = reverse("restclients_proxy", args=[
            "test", "test/v2?i=u\xae_id=A"])
        get_user('test_view')
        self.client.login(username='test_view',
                          password=get_user_pass('test_view'))

        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_format_search_params(self):
        url = 'https://test.edu/api/test?a=one&b=two&c=one%20two'
        self.assertEqual(RestProxyView.format_search_params(url), {
            'a': 'one', 'b': 'two', 'c': 'one two'})

    def test_support_links(self):
        url = reverse("restclients_proxy", args=["test", "test/v1"])
        get_user('test_view')
        self.client.login(username='test_view',
                          password=get_user_pass('test_view'))

        response = self.client.get(url)
        self.assertEquals(response.status_code, 200)

        url = reverse("restclients_proxy", args=[
            "test", "test/v1?a=one&b=two&c=one%20two"])
        response = self.client.get(url)
        self.assertEquals(response.status_code, 200)

    def test_service_errors(self):
        get_user('test_view')
        self.client.login(
            username='test_view', password=get_user_pass('test_view'))

        # Unauthorized service
        url = reverse("restclients_proxy", args=["secret", "test/v1"])
        response = self.client.get(url)
        self.assertEquals(response.status_code, 401)

        # Missing service
        url = reverse("restclients_proxy", args=["fake", "test/v1"])
        response = self.client.get(url)
        self.assertEquals(response.status_code, 404)
        self.assertIn(b"Missing service: fake", response.content)


@override_settings(
    RESTCLIENTS_ADMIN_AUTH_MODULE='rc_django.tests.can_proxy_restclient')
class RestSearchViewTest(TestCase):
    def test_get_context_data(self):
        kwargs = {"service": "libraries", "path": "index.html"}
        context = RestSearchView().get_context_data(**kwargs)
        self.assertEqual(context["wrapper_template"], "proxy_wrapper.html")
        self.assertEqual(context["form_template"],
                         "customform/libraries/index.html")

    def test_search_post(self):
        get_user('test_view')
        self.client.login(username='test_view',
                          password=get_user_pass('test_view'))

        # missing form values
        url = reverse("restclients_customform", args=["libcurrics", "default"])
        response = self.client.post(url)
        self.assertEquals(response.status_code, 400)
        self.assertIn(b"Missing reqired form value: 'campus'",
                      response.content)

        url = reverse("restclients_customform", args=["libcurrics", "default"])
        response = self.client.post(url, {"campus": "sea"})
        self.assertEquals(response.status_code, 302)
        self.assertEqual(
            response.url,
            "/view/libcurrics/currics_db/api/v1/data/defaultGuide/sea")

        url = reverse("restclients_customform", args=["libraries", "accounts"])
        response = self.client.post(url, {"id": "\xae"})
        self.assertEquals(response.status_code, 302)
        self.assertEqual(response.url,
                         "/view/libraries/mylibinfo/v1/?id=%C2%AE")

        url = reverse("restclients_customform", args=["libraries", "accounts"])
        response = self.client.post(url, {"id": "javerage"})
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url,
                         "/view/libraries/mylibinfo/v1/?id=javerage")

    def test_customform(self):
        url = reverse("restclients_customform", args=[
            "libraries", "index.html"])

        get_user('test_view')
        self.client.login(username='no_auth',
                          password=get_user_pass('test_view'))

        # no auth
        response = self.client.get(url)
        self.assertEquals(response.status_code, 302)
        self.assertTrue("next=/search/libraries/index.html" in response.url)

        # with auth
        get_user('test_view')
        self.client.login(username='test_view',
                          password=get_user_pass('test_view'))
        response = self.client.get(url)
        self.assertEquals(response.status_code, 200)
        self.assertIn(b'<form method="post" action="/search/libraries/index">',
                      response.content)

        url = reverse("restclients_customform", args=["fake", "index.html"])
        response = self.client.get(url)
        self.assertEquals(response.status_code, 404)



import os
from setuptools import setup

README = """
See the README on `GitHub
<https://github.com/uw-it-aca/uw-restclients-django-utils>`_.
"""

version_path = 'rc_django/VERSION'
VERSION = open(os.path.join(os.path.dirname(__file__), version_path)).read()
VERSION = VERSION.replace("\n", "")

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

url = "https://github.com/uw-it-aca/uw-restclients-django-utils"
setup(
    name='UW-RestClients-Django-Utils',
    version=VERSION,
    packages=['rc_django'],
    author="UW-IT AXDD",
    author_email="aca-it@uw.edu",
    include_package_data=True,
    install_requires=[
        'Django~=2.2.23',
        'UW-RestClients-Core',
        'django-userservice',
    ],
    license='Apache License, Version 2.0',
    description=('UW-RestClients-Django-Utils'),
    long_description=README,
    url=url,
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
    ],
)
